# Q-matrices size:
Q_MATRIX_ROWS = 15
Q_MATRIX_COLUMNS = 9

# Mass range config:
M_MIN = 0.8         # Minimum mass
CHANDRASEKHAR_LIMIT = 1.4

# Binaries params:
M_SNII = 8.0         # Lower limit for binaries in Supernovae II
B_MIN = 3.0          # Lower limit for binaries in Supernovae I
B_MAX = 16.0         # Upper limit for binaries in Supernovae I
BIN_FRACTION = 0.15  # Fraction of binaries

# Model calculations params:
TOTAL_TIME = 13.25   # Total integration time in Gigayears
